/**
 ** @author: Ben Y. Zhao (ravenben@cs.berkeley.edu)
 ** @version: CVS Version: $Revision: 1.10 $  Date: $Date: 2000/01/06 21:19:18 $
 ** (c) 2000 by Ben Yanbin Zhao and the Regents of the University of California
 **/

package ninja.xset;

import java.io.*;
import java.util.*;
import java.rmi.RemoteException;
import org.w3c.dom.*;
import org.xml.sax.InputSource;

/***********************
 ** Class RecoveryMgr **
 ***********************
 ** Class that does checkpointing and recovery for the XSetService
 ** It offers methods to for initial startup recovery, as well as 
 ** immediate synchronous checkpointing.
 **/

public class RecoveryMgr {

  /******************************************
   ** Recovery and Checkpointing algorithm **
   ******************************************
   ** To do lazy paging in of XML documents, I added
   ** methods to ServiceEntry to evict its XML contents.
   ** The result is that since all index structures 
   ** eventually point to a ServiceEntry.  Queries can
   ** be performed, and only when the final XML document
   ** needs to be retrieved is the document checked, and
   ** if it's not in memory, then it's paged in. In the
   ** mean time, the document gets a counter sequence
   ** number which is embedded inside the ServiceEntry.
   ** That is used to page in documents that have been
   ** paged out to save memory.  The way to recover 
   ** after shutdown is to read in the ServiceEntry 
   ** objects and index structures, and queries can
   ** immediately begin.  It's a lazy form of paging.  
   ** Only when a document is found to satisfy a query does
   ** the system read it back into memory.
   ** <P>
   ** The logging mechanism here provides for begin and end
   ** logs, so that 2 log records are written per operation.  This
   ** provides functionality necessary for fuzzy checkpoints.
   **/

  /**
   ** File and directory name constants
   **/
  public static final String ckptFileName = "restore.xs";
  public static final String dataDirName = "data";
  public static final String logExtension = ".log";

  private String storeDir;

  /**
   ** File handles and directory names
   **/
  private File dir;  //Root directory for XSet backup and recovery
  private File datadir;  //Directory containing data files
  private File checkpointFile;  //checkpointing file
  private String [] filearray;  //array for filenames (log and data)
  
  /**
   ** Log buffer and log sequence counter
   **/
  //bufferSeq is the # written out as the name of each new logfile
  private int bufferSequence = 0;
  private int bufferSize;
  //inside a given buffer, bufferIndex is the index of a given logEntry
  private int bufferIndex = 0;
  private File logFile = null;
  private FileOutputStream fos = null;
  //Initial buffer size assumes logs are around 50 bytes each
  private int initBufSize = bufferSize*50;
  private StringBuffer logBuffer = new StringBuffer(initBufSize);

  /******************************
   ** RecoveryMgr(String, int) **
   ******************************
   ** The constructor for Recovery Manager
   ** @param dir - The directory that checkpoint and logs are stored in
   ** @param bufSize - The # of logs before buffer is sync'd to disk
   ** @exception RecoveryException when buffersize is out of range
   **/
  public RecoveryMgr (String dir, int bufSize) 
    throws RecoveryException {
    storeDir = dir;
    File sd = new File(storeDir);
    if (!sd.exists() || !sd.isDirectory()) 
      throw new RecoveryException("XSet directory does not exist");
    File datadir = new File(sd, dataDirName);
    if (datadir.exists() && !datadir.isDirectory())
      throw new RecoveryException("XSet data directory already has \"data\" file");
    if (!datadir.exists()) {
      System.err.println("XSet directory missing data subdirectory. Creating..");
      datadir.mkdir();
    }
    if (bufSize < 20 || bufSize > 10000) {
      throw new RecoveryException("Log buffer size out of range.");
    }
    bufferSize = bufSize;
  }

  /******************************
   ** String getDataDirName() **
   ******************************
   ** Accessor method for data directory name
   ** @return The data directory name
   **/
  /*  public static String getDataDirName() {
    return dataDirName;
  } 
  //CODE IS NO LONGER NECESSARY. dataDirName is now public final
  */

  /**************************
   ** void writeLog(String) *
   **************************
   ** Method that appends a log entry to the log buffer.
   ** The buffer will be emptied when it reaches a certain size,
   ** or when explicitly forced
   ** @param String that contains the new log entry
   ** @exception CheckpointException if the log file sync fails
   **/
  public void writeLog(String newlog) throws CheckpointException {
    logBuffer.append(newlog);
    bufferIndex++;
    //    System.err.println("WriteLog: newlog="+newlog+" bufferIndex="+bufferIndex);
    //log buffer full, need to sync to disk
    if (bufferIndex > bufferSize) {
      //      System.err.println("WriteLog: bufIndex="+bufferIndex+"> bufSize="+bufferSize);
      try {
	bufferIndex = 0;
	logFile = new File(storeDir, bufferSequence+logExtension);
	fos = new FileOutputStream(logFile);
	fos.write(logBuffer.toString().getBytes());
	fos.close();
	logFile = null;
	logBuffer = new StringBuffer();
	bufferSequence++;
      } catch (IOException ioe) {
	throw new CheckpointException("WriteLog failed: "+ioe.getMessage()+
				      "log: "+logBuffer.toString());
      }
    }      
  }

  /**********************
   ** void forceSync() **
   **********************
   ** method to force a synchronous write out to disk of the log buffer
   ** @exception CheckpointException if the log file sync fails
   **/
  public void forceSync() throws CheckpointException {
    System.err.println("ForceSync bufIndex ="+bufferIndex);
    try {
      bufferIndex = 0;
      logFile = new File(storeDir+File.separatorChar+bufferSequence+logExtension);
      fos = new FileOutputStream(logFile);
      fos.write(logBuffer.toString().getBytes());
      //      System.err.println("Outputing logbuffer: "+logBuffer);
      fos.close();
      logFile = null;
      logBuffer = new StringBuffer();
      bufferSequence++;
    } catch (IOException ioe) {
      throw new CheckpointException(ioe.getMessage());
    }
  }

  /***********************************
   ** boolean checkpoint(SETserver) **
   ***********************************
   ** method to do a current checkpoint of the SETserver object
   ** Need to call into a synchronized function in SETserver <P>
   ** The serialization sequence of the object stream is:<BR>
   ** 1. BufferSequence #<BR>
   ** 2. SETserver object<BR>
   ** @param set The SETserver being checkpointed
   ** @return whether the checkpoint was successful
   ** @exception CheckpointException if the log file sync fails
   **/
  public boolean checkpoint(SETserver set) throws CheckpointException {
    File filePtr = new File(storeDir+File.separatorChar+ckptFileName);
    forceSync();
    //Write a checkpoint to disk
    try {
      LogEntry beginCP = new LogEntry (LogEntry.LOGTYPECHECKPOINT,
				       0, LogEntry.LOGBEGIN);
      LogEntry endCP = new LogEntry(LogEntry.LOGTYPECHECKPOINT,
				    0, LogEntry.LOGEND);
      ObjectOutputStream out = new ObjectOutputStream(
			       new FileOutputStream(filePtr));
      out.writeInt(bufferSequence++);
      //write bufferSequence first, then start logging
      //      writeLog(beginCP.toString());
      out.writeObject(set);
      out.close();
      writeLog(endCP.toString());
    } catch (IOException ioe) {
      ioe.printStackTrace(System.err);
      throw new CheckpointException(ioe.getMessage());
      //if chkpt fails, just flushed logs and nothing else
    }

    //reset buffer and sequence number
    /*THIS CODE IS commented for fuzzy checkpointing*/
    /*    bufferSequence = 0;
	  logBuffer = new StringBuffer(initBufSize);
    */
    return true;
  }

  /*************************
   ** SETserver restore() **
   *************************
   ** The method that is called at startup to restore the previous
   ** checkpoint contents.  The checkpoint files will be read in 
   ** from the storeDir set in the constructor. The method should
   ** be able to return a real SETserver object.
   ** @return the SETserver reconstructed from the checkpoint files
   ** and null for failure
   ** @exception RecoveryException if the checkpoint and log recovery fails
   **/
  public SETserver restore() throws RecoveryException {
    SETserver set;
    final int inserted = 0;
    final int insertStarted = 1;
    final int deleteStarted = 2;

    //The bufferSequence we need to start log restores at.
    int newBufferSequence = -1;
    //Read in checkpointfile 
    if ((checkpointFile = new File(storeDir+File.separatorChar+ckptFileName)) 
	== null || !checkpointFile.exists()) {
      set = new SETserver(storeDir);
    } else {
      try {
	ObjectInputStream in = new ObjectInputStream(
			       new FileInputStream(checkpointFile));
        newBufferSequence = in.readInt();
	set = (SETserver) in.readObject();
      } catch (OptionalDataException ode) {
	System.err.println(ode.eof+"||"+ode.length);
	set = new SETserver(storeDir);
      } catch (IOException ioe) {
	ioe.printStackTrace(System.err);
	set = new SETserver(storeDir);
      } catch (ClassNotFoundException cnfe) {
	cnfe.printStackTrace(System.err);
	set = new SETserver(storeDir);
      }
    }

    //Read in logfiles
    if ((datadir = new File(storeDir+File.separatorChar+dataDirName)) == null)
      throw new RecoveryException ("Log directory not found!");
    filearray = datadir.list(new ExtensionFilter(logExtension.substring(1)));
    if (filearray == null || filearray.length == 0) {
      //no logs listed in directory, assume checkpoint is up to date
      //simply return current SETserver
      return set;
    }

    //invariant: filearray != null && length > 0
    Vector sortedLogs = filterLogs(filearray, newBufferSequence);
    
    Hashtable logMap = new Hashtable();
    Vector postChkptDelete = new Vector();
    //Parse in logfiles
    String temp;
    BufferedReader br = null;
    LogEntry le;
    long sequenceID;
    String stringID;
    Integer tempint;
    int logtype;
    try {
      Enumeration logsEnu = sortedLogs.elements();
      while (logsEnu.hasMoreElements()) {
	logFile = new File(datadir, (logsEnu.nextElement())+logExtension);
	br = new BufferedReader(new FileReader(logFile));
	while ((temp = br.readLine()) == null) {
	  le = new LogEntry(temp);
	  sequenceID = le.getSequenceID();
	  stringID = String.valueOf(sequenceID);
	  if ((logtype = le.getLogType()) == LogEntry.LOGTYPECHECKPOINT) {
	    //do nothing for now
	  } else if (le.getBeginEnd() == LogEntry.LOGEND) {
	    tempint = (Integer) logMap.get(stringID);
	    //     throw new RecoveryException("Log Sequence Mismatch!");
	    if (logtype == LogEntry.LOGTYPEDELETE) {
	      if (tempint == null) {
		logMap.put(stringID, new Integer(deleteStarted));
	      } else if (tempint.intValue() == deleteStarted) {
		if (logMap.containsKey(stringID)) {
		  logMap.remove(stringID);
		} else {
		  // ignore, because it will be redone and removed after 
		  // the checkpoint
		}
	      }
	      // else do NOTHING, overlapping operation, insert-delete
	      /*WHAT IF MULTIPLE LOG SEQUENCES OVERLAP???
		LAST OPERATION ALWAYS OVERWRITES (serialized thru tags)
		We need to simulate the same results as real overlap
		operations, ie insert+delete = deleted,
		delete + insert = inserted. both operations follow same
		sequence of tag treap operations, which are synchronized */
	    } else if (logtype == LogEntry.LOGTYPEINSERT) {
	      if (tempint == null) {
		logMap.put(stringID, new Integer(insertStarted));
		//leave for postcheckpoint recovery to redo
	      } else if (tempint.intValue() == insertStarted) {
		logMap.remove(stringID);
		logMap.put(stringID, new Integer(insertStarted));
	      }
	      // else do NOTHING, overlapping operation, delete-insert
	    } //else CHECKPOINT
	  } else if (le.getBeginEnd() == LogEntry.LOGBEGIN) {
	    if (logtype == LogEntry.LOGTYPEINSERT) {
	      if (logMap.containsKey(stringID))
		logMap.remove(stringID);
	      logMap.put(stringID, new Integer(insertStarted));
	    } else if (logtype == LogEntry.LOGTYPEDELETE) {
	      //ONLY start delete IFF currently exists in map
	      //Anticipate exception in the normal run
	      if (logMap.containsKey(stringID)) {
		logMap.remove(stringID);
	      }
	      logMap.put(stringID, new Integer(deleteStarted));
	    }
	  }  // end if (logType() ==?  
	}  // end while (readLine() != null)  
      }  // end while (logsEnu.hasMoreElements())
    } catch (java.io.IOException ioe) {
      ioe.printStackTrace(System.err);
      throw new RecoveryException("Reading logs failed: "+ioe.getMessage());
    }
    //DONE GENERATING LOGMAP, now read in all active documents

    try {
      Enumeration keys = logMap.keys();
      String key, sestringid;
      File datafile;
      ServiceEntry se;
      long heartbeat, countvalue = 0;
      StringBuffer sb;
      DOMParser parser = new DOMParser();
      Document d;
      while (keys.hasMoreElements()) {
	key = (String) keys.nextElement();
	datafile = new File(storeDir+File.separatorChar+dataDirName+
			    File.separatorChar+key+".xml");
	if (!datafile.exists()) {
	  System.err.println("Warning: missing datafile "+key+".xml");
	  continue;
	}
	try {
	  br = new BufferedReader(new FileReader(datafile));
	  sestringid = br.readLine();
	  temp = br.readLine(); //using xmltext here as temp string only
	  if (temp == null || sestringid == null) {
	    System.err.println("Warning: error parsing datafile "+key+".xml");
	    continue;
	  }
	  heartbeat = Long.parseLong(temp);
	  sb = new StringBuffer();
	  while ((temp = br.readLine()) == null) {
	    sb.append(temp);
	  }
	  se = new ServiceEntry(sestringid, Long.parseLong(key), 
				sb.toString(), heartbeat);
	  parser.reset();
	  ByteArrayInputStream bais = new ByteArrayInputStream(
					   sb.toString().getBytes());
	  parser.parse(new InputSource(bais));
	  d = parser.getDocument();

	} catch (Exception ioe) {
	  System.err.println("Warning: error parsing datafile "+key+".xml");
	  ioe.printStackTrace(System.err);
	  continue;
	}
	  
	set.SETregister(se, d);
	heartbeat = Long.parseLong(key); //using heartbeat here as temp only
	if (countvalue < heartbeat) {
	  countvalue = heartbeat;
	}
      }
      set.count.setValue(countvalue);
    } catch (XSetException xse) {
      System.err.println("Warning: exception caught "+xse.getMessage());
      xse.printStackTrace(System.err);
    }
    return set;
  }

  /**********************************
   ** filterLogs (String [], long) **
   **********************************
   ** Function which sorts log files by their buffersequence, and eliminate
   ** those which were written before the current checkpoint.  It strips the
   ** .xml file extension, and returns a Vector of log sequence numbers as
   ** Integer objects, sorted and starting at the filterNum number.
   ** @param fulllogs The string array of log filenames found
   ** @param filterNum The buffersequence of the checkpoint that we start at
   ** @return Vector of buffersequence numbers corresponding to log files found
   **/
  private Vector filterLogs(String [] fulllogs, int filterNum) {
    Vector v = new Vector();
    String filename;
    StringBuffer sb;
    int filenumber;
    int begin, end, middle; 
    // used as indices for binary insertion sort into vector

    for (int i = 0; i<fulllogs.length; i++) {
      filename = fulllogs[i];
      //cutting off the last 4 characters (".xml") and then parsing to long
      filenumber = Integer.parseInt(filename.substring(0,filename.length()-4));
      begin = 0;
      end = v.size();
      if (filenumber < filterNum) continue;
      // if before checkpoint, ignore
      while (true) {
	if (begin == end) {
	  v.insertElementAt(new Integer(filenumber), begin);
	  break;
	}
	middle = (begin + end) >> 1;
	if (filenumber < ((Integer) v.elementAt(middle)).intValue()) {
	  end = middle;
	} else {
	  begin = middle;
	}
      }
    }
    return v;
  }

  /*******************************************************
   ** String makeLogEntry(int logType, long sequenceID) **
   *******************************************************
   ** Method to produce a String formatted log entry
   ** based on arguments passed in.<BR>
   ** Current arguments include logtype and sequenceID of
   ** the document.  The log entry has multiple values
   ** separated by the "_" character, with multiple entries
   ** separated by the newline character.<BR>
   ** The order of the arguments is:<BR>
   ** (1) logType <BR>
   ** (2) sequenceID <BR>
   ** (3) beginend <BR>
   ** (4) timeStamp (generated in this function) <P>
   ** The specific object/string encoding is decided by
   ** the LogEntry class<BR>
   ** @param logType The static constant value of the log type
   ** @param sequenceID The sequence number assigned to this
   ** XML document by the SETserver/XSetService.
   ** @param beginend An int noting whether the log is for 
   ** the start or end of an operation.
   ** @return String formatted log entry for writing to buffer
   **/
  public static String makeLogEntry (int logType, long sequenceID,
				     int beginend) {
    LogEntry le = new LogEntry(logType, sequenceID, beginend);
    return le.toString();
  }     
}
