/***
 ** @author: Ben Y. Zhao (ravenben@cs.berkeley.edu)
 ** @version: CVS Version: $Revision: 1.15 $  Date: $Date: 2000/01/06 21:19:18 $
 ** (c) 2000 by Ben Yanbin Zhao and the Regents of the University of California
 ** 
 ** This is the code release of the XSet XML search engine on treaps.  The XSet
 ** uses treaps to represent hierarchies of tags and indices on tag values.  
 ** An ongoing progress homepage of the project can be found at
 ** http://www.cs.berkeley.edu/~ravenben/xset
 ** SETserver.java is the main Server code for XSet
 ** <P>
 ** The 1.1 release moves from the Microsoft MSXML 1.9 parser to the XML4J
 ** parser 2.0.9.
 **/

package ninja.xset;

import java.util.*;
//import ninja.*;
import java.io.*;
import org.w3c.dom.*;
import org.xml.sax.InputSource;

/**
 ** Class that provides the main interface to the Search Engine on Treaps
 ** It provides the public methods register and deregister for the
 ** external applications.
 */
public class SETserver implements Serializable {
  
  public static final short HASHSTORE = 0;
  public static final short TREAPSTORE = 1;
  
  protected Hashtable rootTagList;
  protected boolean refresh = false;
  protected long timeout = 0;
  protected Hashtable globalHashSID = null;
  protected String storeDir = null;
  Counter count = null;  //allow xset code to reset or define counter

  protected transient Thread runCleaner = null;
  protected transient DOMParser parser = null;

  /** TODO: insert abstraction layer between hashtables and serviceEntries  
   **   This is sort of done for now, with the evict and pagein methods
   **   added to ServiceEntry.  But nothing uses those methods yet
   **/

  /*********************
   ** SETserver(long) **
   *********************
   ** Constructor method
   ** Initializes variables and prepares the search indices for data
   ** @param refreshRate The number of milliseconds taken for an entry to become outdated encoded as a long
   ** @param sDir The storage directory for all XML files
   **/
  public SETserver(long refreshRate, String sDir) {
    rootTagList = new Hashtable();
    globalHashSID = new Hashtable();
    parser = new DOMParser();
    storeDir = sDir;
    count = new Counter();
    if (refreshRate > 0) {
      timeout = refreshRate;
      refresh = true;
      runCleaner = new Thread(new SETcleaner(this, timeout));
      runCleaner.start();
    } else {
      System.err.println("ERROR: refreshRate < 0!");
      System.exit(-1);
    }
  }
  
  /*****************
   ** SETserver() **
   *****************
   ** Constructor method
   ** Initializes variables and prepares the search indices for data
   ** This constructor is for SET instances with no timeouts
   ** @param sDir The storage directory for all XML files
   */
  public SETserver(String sDir) {
    rootTagList = new Hashtable();
    globalHashSID = new Hashtable();
    parser = new DOMParser();
    storeDir = sDir;
    count = new Counter();
  }
  
  /**********************************
   ** boolean containsSEID(String) **
   **********************************
   ** Method that tests if a given ServiceEntry ID is already registered 
   ** in SETserver.
   ** @return whether the hashtable contains the given ServiceEntry ID
   **/
  public boolean containsSEID(String serviceID) {
    return globalHashSID.containsKey(serviceID);
  }

  /*******************************
   ** Vector querySET(Document) **
   *******************************
   ** Method called by the XSet service or independent application to
   ** query for a certain document, given an XML template
   **
   ** @param query The XML document used as the template for the search
   ** @return The serviceEntry vector containing (if any) the matches
   **  that correspond to the query
   **/
  public Vector querySET(Document query) {
    //    System.err.println("Entering querySET(Doc, SETName): "+SETName);
    Hashtable rootHash = null;
    Treap rootTreap = null;
    if (query == null) return new Vector();
    Element root = query.getDocumentElement();
    HashVector results = null;
    int i = SETutils.numChildren(root);
    String SETName = root.getTagName();
    //NOTE: replace & with whatever character is not allowed in tag names
    short x;
    if (i>1 || (i==1 && 
		(root.getFirstChild().getNodeType()==Node.ELEMENT_NODE))) {
      rootHash = (Hashtable) rootTagList.get(SETName+"&"+SETserver.HASHSTORE);
      // NEVER return null, always return empty Vector
      //System.err.println("querying hash");
      if (rootHash == null) return new Vector();
      results = queryTag(rootHash, root);
    } else if (i==1 && 
	       ((x = root.getFirstChild().getNodeType())==Node.TEXT_NODE
		|| x == Node.CDATA_SECTION_NODE)) {
      rootTreap = (Treap) rootTagList.get(SETName+"&"+SETserver.TREAPSTORE);
      // NEVER return null, always return empty Vector
      if (rootTreap == null) return new Vector();
      //System.err.println("querying hash");
      results = queryTag(rootTreap, root);
    } else {
      System.err.println("ERROR: Root element contains empty or invalid query!");
      return new Vector();
    }
    return totalIntersect(results);
  }
  
  /*******************************
   ** querySET(Document, Vector)**
   *******************************/
  /** Method to query the XSet server, and return only those fields requested
   ** <P>
   ** This returned data structure is rather complicated.  It's a vector of
   ** hashtables, one hashtable for each matching XML document found. For
   ** each of these matches, the hashtable hashes each of the requested
   ** tag names to a vector.  The vector contains all the string values
   ** which occur in that tagname in the currently matched XML document.
   ** So it's a Vector of Hashtables of Vectors.
   ** <P>
   ** @param xmlQuery, the xml query in the form of a DOM tree object
   ** @param tagVector, a vector containing String names of requested tags
   ** @return Vector of hashtables that map each requested tagname to its value
   ** @exception XSetException for normal XSet errors
   **/
  public Vector querySET(Document xmlQuery, Vector tagVector)
    throws XSetException {
    Vector results;
    results = querySET(xmlQuery);
    if (results.size() == 0) return results;
    Vector returnVector = new Vector();
    Enumeration docs = results.elements();
    Enumeration tagNames;
    ServiceEntry se;
    Document matchingDoc;
    String matchingXML, tagName;
    ByteArrayInputStream bais;
    int debug = 0;
    while (docs.hasMoreElements()) {
      // FOR each matching Document, DO:
      parser.reset();
      se = (ServiceEntry) docs.nextElement();
      matchingXML = se.getXML(storeDir+File.separatorChar+
			      RecoveryMgr.dataDirName);
      System.err.println("Document # "+(++debug)+"\n"+matchingXML);
      bais = new ByteArrayInputStream(matchingXML.getBytes());
      try {
	parser.parse(new InputSource(bais));
      } catch (Exception e) {
	e.printStackTrace(System.err);
      }
      matchingDoc = parser.getDocument();
      Hashtable ht = new Hashtable();
      tagNames = tagVector.elements();
      while (tagNames.hasMoreElements()) {
	// FOR each requested tagname, DO:
	tagName = (String) tagNames.nextElement();
	System.err.println("Document #"+debug+", tagName:"+tagName);
	NodeList nl = matchingDoc.getElementsByTagName(tagName);
	Vector v = new Vector();
	Element e;
	for(int i=0; i < nl.getLength(); i++) {
	  // FOR each tag found of this tagname, DO:
	  System.err.println("Document #"+debug+
			     ", found tagNm:"+tagName);
	  v.addElement(getText(nl.item(i)));
	} //for
	ht.put(tagName, v);
      } // while
      returnVector.addElement(ht);
    }
    return returnVector;
  }      
  private String getText(Node n) {
    Node child;
    if (n.getNodeType() == Node.ELEMENT_NODE) {
      
      if ((child = n.getFirstChild()) != null) {
	return child.getNodeValue();
      }
    } else if (n.getNodeType() == Node.TEXT_NODE)
      return n.getNodeValue();
    return null;
  }
  
  /***************************************
   ** Enumeration enumerateTag (Vector) **
   ***************************************
   ** Method to return an enumeration of values for a given tag.
   ** The method traverses the internal index for tags, and upon finding
   ** the desired tag (indicated through a series of string arguments, it
   ** calls the keys() method on the treap, and returns the enumeration.
   ** @param pathArray is an array of strings that denote the path from
   ** the root of the document to the desired tag
   ** @return an enumeration of the existing values for the specified given
   ** tag in the dataset tree
   ** @exception XSetException for normal XSet errors
   **/
  public Enumeration enumerateTag(String [] pathArray) 
    throws XSetException {
    Treap tagTreap;
    if (pathArray == null) 
      throw new XSetException("Null input to enumerateTag!");
    String tagName;
    Vector newV = new Vector();
    Enumeration enu;
    if (pathArray.length == 1) {
      tagTreap = (Treap) rootTagList.get(pathArray[0]+"&"+
					 SETserver.TREAPSTORE);
      if (tagTreap == null) return (new Hashtable()).keys();
      return tagTreap.keys();
    }
    Hashtable tagHash;
    tagHash = (Hashtable) rootTagList.get(pathArray[0]+"&"+
					  SETserver.HASHSTORE);
    int i = 0, j = pathArray.length - 1;
    // grab the last value, that's the index to the Treap, not Hashtable
    try {
      for (i = 1; i < j; i++) {
	System.out.println("i: "+pathArray[i]+" j: "+pathArray[j]);
	// iterate from the 2nd value to the next to last value
	tagHash = (Hashtable) tagHash.get(pathArray[i]+"&"+
					  SETserver.HASHSTORE);
	if (tagHash == null) return (new Hashtable()).keys();
      }
      System.out.println(pathArray[j]);
      tagTreap = (Treap) tagHash.get(pathArray[j]+"&"+
				     SETserver.TREAPSTORE);
    } catch (ClassCastException cce) {
      // illegal cast, means we hit the treap faster than we expected
      // so we just cast it to the treap, cross fingers and return values
      tagTreap = (Treap) tagHash.get(pathArray[i]+"&"+
				     SETserver.TREAPSTORE);
    }
    return tagTreap.keys();
  }
  
  /****************************************
   ** Vector totalIntersect (HashVector) **
   ****************************************
   ** Method to do a group intersection of the total hashtables returned
   ** by each tag search
   ** TESTED!!
   **/
  //XML4J CONVERSION NOT NECESSARY
  private Vector totalIntersect(HashVector hv) {
    //    System.err.println("Entering totalIntersect(hv)");
    if (hv == null) {
      System.err.println("ERROR: Unexpected empty HashVector in Intersect");
      return new Vector();
    }
    if (!hv.check()) {
      System.err.println("ERROR: consistency check in hashVector failed!");
      return new Vector();
    }
    Vector results = new Vector();
    if (hv.noresult) return results; // short circuit
    Hashtable counter = new Hashtable();
    Hashtable stored = new Hashtable();
    
    Enumeration hashtables = hv.hashtables.elements();
    Enumeration keys;
    String key;
    Hashtable hPtr;
    Integer count;
    while (hashtables.hasMoreElements()) {
      hPtr = (Hashtable) hashtables.nextElement();
      keys = hPtr.keys();
      while (keys.hasMoreElements()) {
	key = (String) keys.nextElement();
	stored.put(key, hPtr.get(key));
	count = (Integer) counter.get(key);
	if (count == null) count = new Integer(1);
	else {
	  hPtr.remove(key);
	  count = new Integer(count.intValue() +1);
	}
	counter.put(key, count);
      }
    }
    keys = counter.keys();
    int i ;
    while (keys.hasMoreElements()) {
      key = (String) keys.nextElement();
      i = ((Integer) counter.get(key)).intValue();
      if (i == hv.size) {
	results.addElement(stored.get(key));
      }
    }
    return results;
  }
  
  /*********************
   ** Class textUnion **
   *********************
   ** Class to encapsulate the text and children tags
   ** Right now it's not used, because the parser functionality is not there.
   ** Tried to implement this and do a substring query on the current 
   ** getText(), but that doesn't work, since substring does not allow for
   ** missing words, etc etc.
   **/
  /*  private class textUnion {
      public String text;
      public Hashtable children;
      public textUnion(String t, Hashtable h) {
      text = t;
      children = h;
      }
      }
  */
  
  /************************
   ** Class HashVector() **
   ************************
   ** Class to keep track of multiple responses in hashtables.
   ** Use this to return list of hashtables that we get from
   ** range queries, and also to combine recursive results into larger sets
   **/
  //XML4J CONVERSION NOT NECESSARY
  private class HashVector {
    public int size;
    public Vector hashtables;
    boolean noresult;
    public HashVector(Hashtable h) {
      if (h == null) {
	noresult = true;
	size = 0;
      } else {
        noresult = false;
	size = 1;
	hashtables = new Vector();
	hashtables.addElement(h);
      }
    }
    public HashVector() {
      noresult = false;
      size = 0;
      hashtables = new Vector();
    }
    public HashVector(boolean shortcircuit) {
      noresult = true;
      size = 0;
    }
    //function to check consistency inside the hashVector
    public boolean check() {
      if (noresult == false) return (size == hashtables.size());
      return true;
    }
    public void addHashtable(Hashtable h) {
      size++;
      hashtables.addElement(h);
    }
    /**
     * Adds the contents of hashVector hv to this hashVector
     */
    public void merge(HashVector hv) {
      if (hv.noresult == true || noresult == true) {
	noresult = true;
	return;
      }
      Enumeration e = hv.hashtables.elements();
      while (e.hasMoreElements()) {
	this.addHashtable((Hashtable) e.nextElement());
      }
    }
  }
  
  /*****************************************
   ** HashVector queryTag(Treap, Element) **
   *****************************************
   ** Overloaded method that does a queryTag on an end node.  The data 
   ** structure in question is a Treap.
   ** @param t The treap that we're using to match up against the element
   ** @param e The query node that should match the treap of values
   ** @return A HashVector object that contains the hashtables
   **/
  private HashVector queryTag(Treap t, Element e) {
    if (t == null || e == null) {
      System.err.println("Null values in queryTag. Please contact author!");
      return new HashVector();
    }
    //    System.err.println("Entering queryTag(Treap, Element): "+
    //		       e.getTagName().toString()+" "+e.getText());
    HashVector hv = null;
    Hashtable h = null;
    //This should be an end node, hence no ELEMENT children, only TEXT
    //    Ninja.assert(e.numElements() == 1);
    
    TreapOper to = null;
    Attr att = null;
    //operator values includes whether range is inclusive
    //and the type of the value used as the key
    if ((att = (Attr) e.getAttributeNode("XSetLT")) != null) {
      //      System.out.println("U: DOING RANGE LOOKUP");
      to = new TreapOper();
      to.lessThan = att.getValue();
      to.lessEquals = false;
      e.removeAttributeNode(att);
    }
    if ((att = (Attr) e.getAttributeNode("XSetMT")) != null) {
      //      System.out.println("V: DOING RANGE LOOKUP");
      if (to == null) to = new TreapOper();
      to.moreThan = att.getValue();
      to.moreEquals = false;
      e.removeAttributeNode(att);
    }
    if ((att = (Attr) e.getAttributeNode("XSetLE")) != null) {
      //      System.out.println("W: DOING RANGE LOOKUP");
      if (to == null) to = new TreapOper();
      to.lessThan = att.getValue();
      to.lessEquals = true;
      e.removeAttributeNode(att);
    }
    if ((att = (Attr) e.getAttributeNode("XSetME")) != null) {
      //      System.out.println("Y: DOING RANGE LOOKUP");
      if (to == null) to = new TreapOper();
      to.moreThan = att.getValue();
      to.moreEquals = true;
      e.removeAttributeNode(att);
    }
    if (to != null) {  // Range query exists
      if (to.lessThan == null && to.moreThan == null) {
	System.err.println("WARNING: range query has no bounds");
	//ASSUMPTION: if it is a leaf node, then it has 1 TEXT CHILD
	h = (Hashtable) t.lookup(e.getFirstChild().getNodeValue());
	return new HashVector(checkAttributes(h, e));
      }
      att = (Attr) e.getAttributeNode("XSetKTYPE");
      if (att.getValue().equalsIgnoreCase("STRING"))
	to.keyType = TreapOper.STRINGKEY;
      else if (att.getValue().equalsIgnoreCase("INTEGER"))
	to.keyType = TreapOper.INTKEY;
      else if (att.getValue().equalsIgnoreCase("FLOAT"))
	to.keyType = TreapOper.FLOATKEY;
      else System.err.println("ERROR: Invalid rangeQuery KeyType!");
      e.removeAttributeNode(att);
      Vector tempV = t.rangeLookup(to); // Do range query
      //      System.out.println("Z: FINISHED RANGE LOOKUP");
      Enumeration enu = tempV.elements();
      Vector addHash = new Vector();
      while (enu.hasMoreElements()) {
	h = (Hashtable) enu.nextElement();
	addHash.addElement(h);
	// check attributes and then add to HashVector
      }
      // if no results after checkAttributes, then short circuit
      if ((h = checkAttributes(combineHash(addHash), e)) == null) {
	hv = new HashVector(false);
      } else {
        hv = new HashVector(h);
      }
    } else {   // no range queries, simple lookup
      h = (Hashtable) t.lookup(e.getFirstChild().getNodeValue());
      // if no results after checkAttributes, then short circuit
      if ((h = checkAttributes(h, e)) == null) {
	hv = new HashVector(false);
      } else {
	hv = new HashVector(h);
      }
    }
    return hv;
  }
  
  /*************************************
   ** Hashtable combineHash(Vector v) **
   *************************************
   ** Method that combines the contents of a bunch of hashtables
   ** @param v the vector containing the hashtables to be combined
   ** @return one hashtable that contains all entries taken from the vector
   **/
  private Hashtable combineHash(Vector v) {
    if (v == null) {
      System.err.println("Warning: null vector in CombineHash.");
      return new Hashtable();
    }
    Enumeration e = v.elements(), keys, hashes;
    Hashtable total = new Hashtable(), h;
    String key;
    while (e.hasMoreElements()) {
      h = (Hashtable) e.nextElement();
      keys = h.keys();
      while (keys.hasMoreElements()) {
	key = (String) keys.nextElement();
	total.put(key, h.get(key));
      }
    }
    return total;
  }
  
  /*********************************************
   ** HashVector queryTag(Hashtable, Element) **
   *********************************************
   ** Method to do the recursive XMLTag treap traversal
   ** This only takes care of the recursive case.
   ** NOTE: This isn't great, since it does non-end recursion
   **
   ** @param t The treap that we're using to matching up against the element
   ** @param e The element that should be matched up against the treap
   ** @return A HashVector with the hashtable of serviceEntries that 
   **         match on the query
   **/
  private HashVector queryTag(Hashtable topHash, Element e) {
    //    System.err.println("Entering queryTag(Hash, Element): "+
    //		       e.getTagName().toString()+" "+e.getText());
    if (topHash == null || e == null) {
      System.err.println("Warning: null input in QueryTag(hash)");
      return new HashVector();
    }
    Hashtable hPtr = null;
    Treap tPtr = null;
    HashVector hv = new HashVector();
    Element elem = null;
    Element currentE = null;
    String tagName;
    
    Vector ec = SETutils.getChildren(e);
    int j = ec.size();
    
    for (int i = 0; i < j; i++) {
      if (hv.noresult) return hv; // if noresult then short circuit
      currentE = (Element) ec.elementAt(i);
      tagName = currentE.getTagName();
      if (currentE.getChildNodes().getLength() == 1 &&
	  currentE.getFirstChild().getNodeType() != Node.ELEMENT_NODE) {
	tPtr = (Treap) topHash.get(tagName+"&"+SETserver.TREAPSTORE);
	if (tPtr != null) hv.merge(queryTag(tPtr, currentE));
	else {
	  // No documents found, so no match possible,
	  // short circuit the whole hashvector to give quick null result
	      hv.merge(new HashVector(false));
	}
      } else {
	hPtr = (Hashtable) topHash.get(tagName+"&"+SETserver.HASHSTORE);
	hv.merge(queryTag(hPtr, currentE));
      }
    }
    return hv;
    
    /* Here's a PROBLEM!!! currently no solution as to how to handle
     * attribute tags for matching attributes for nodes with children
     * elements.
     * TEMP. SOLN.: checkAttributes in queryTag(TREAPS, ELEMENTS)
     */
  }

  /***************************************************
   ** Hashtable checkAttributes(Hashtable, Element) **
   ***************************************************
   ** Method to check that all XML document entries in this Hashtable match
   ** the attribute(s) that we are looking for.
   ** @param v Vector that contains the documentEntries
   ** @param e Element that we're matching attributes up against
   ** @return a hashtable that contains the serviceEntrys which match
   **/
  private Hashtable checkAttributes(Hashtable v, Element e) {
    //  System.err.println("Entering chckAttr(Hash, Element): "+e.getText());
    Hashtable ansV = new Hashtable();
    if (v == null) return (Hashtable) null;
    if (e == null) {
      System.err.println("Warning: null element input to chkAttr.");
      return (Hashtable) null;
    }
    Enumeration docKeys = v.keys();
    DocumentEntry de = null;
    Document d = null;
    
    // put the Element attributes key value pairs into a hashtable
    Hashtable eleAttrs = new Hashtable();
    NamedNodeMap ele = e.getAttributes();
    
    /*    int attrnumber = ele.getLength();
	  for (int i = 0; i < attrnumber; i++) {
	  a = (Attribute) ele.item(i);
	  System.err.println("attribute "+a.getName().toString()+" v: "+a.getValue());
	  eleAttrs.put(a.getName().toString(), a.getValue());
      }*/
    
    Hashtable temphash = new Hashtable();
    Hashtable attrs = null;
    boolean matched;
    String currKey = null;
    while (docKeys.hasMoreElements()) {
      matched = true;
      currKey = (String) docKeys.nextElement();
      de = (DocumentEntry) v.get(currKey);
      attrs = de.attributes;
      if (ele.getLength() == 0  // if query has no attr, shortcut return
	  || nnmsubset(ele, attrs)) {
	ansV.put(currKey, de.se);
      }
    }
    if (ansV.isEmpty()) return (Hashtable) null;
    return ansV;
  }

  /*************************************************
   ** boolean nnmsubset (NamedNodeMap, Hashtable) **
   *************************************************
   ** Method to do strict subset comparison between NamedNodeMaps
   ** @param a NamedNodeMap
   ** @param b Hashtable
   ** @return whether namedNodeMap a is a subset of Hashtable b
   **/
  private boolean nnmsubset(NamedNodeMap a, Hashtable b) {
    if (a == null) return true;
    if (b == null) return false;
    int j = a.getLength();
    Attr att = null;
    if (b.size() < j) return false; // if size < ele, shortcut return
    for (int i = 0; i < j; i++) {
      att = (Attr) a.item(i);
      if (!att.getValue().equals(b.get(att.getName())))
	return false; // if one attr mismatch, then shortcut
    }
    return true;
  }
  
  /*************************
   ** SETrenew (Document) **
   *************************
   ** Method called by he SDS server to update an existing 
   ** XML document.  It hashes the document to get the ID and
   ** retrieves the serviceEntry, then updates the heartbeat field.
   ** @param renewDoc XML document to be renewed
   ** @exception XSetException for generalized errors
   **/
  public void SETrenew(Document d) throws XSetException {
    if (d == null) return; // null document, no renewal necessary
    String serviceID = null;
    ServiceEntry se = null;
    try {
      serviceID = SETutils.hasher(d, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace(System.err);
      throw new XSetException("Missing SHA-1 hashing algorithm!");
    }
    se = (ServiceEntry) globalHashSID.get(serviceID);
    se.updateTime(System.currentTimeMillis());
  }
  
  /************************************
   ** SETreplace(Document, Document) **
   ************************************
   ** Method called by the SDS server or independent application to
   ** replace an existing document with a new document.  It deletes
   **e old document first, the registers the new document.
   ** @param olddoc The XML document to be replaced in the SET
   ** @param newdoc The XML document to replace olddoc
   ** @exception XSetException for generalized errors
   **/
  public void SETreplace(Document olddoc, Document newdoc) 
    throws XSetException {
    if (olddoc == null || newdoc == null) 
      throw new XSetException("Null documents for SETreplace.");
    SETderegister(olddoc);
    SETregister(newdoc, count.getNextCount());
  }
  
  /*******************************
   ** SETregister(String, long) **
   *******************************
   ** Method called by application to add new document to XSet.
   ** This version is to provide a wrapper for parsing the XML string
   **
   ** @param xmltext The XML document in string format
   ** @param counter The counter value assigned externally via a counter obj
   ** @exception XSetException for generalized errors
   **/
  public void SETregister(String xmltext, long counter) 
    throws XSetException {
    //    System.err.println("Entering SETregister(String: "+xmltext+")");
    if (xmltext == null)
      throw new XSetException("Null string input for SETregister");
    Document d = null;
    try {
      parser.reset();
      ByteArrayInputStream bais = new ByteArrayInputStream(xmltext.getBytes());
      parser.parse(new InputSource(bais));
      d = parser.getDocument();
    } catch (Exception e) { 
      e.printStackTrace(System.err); 
      throw new XSetException(e.getMessage());
    }
    if (d == null) throw new XSetException("Error in parsing: "+xmltext);
    ServiceEntry se = new ServiceEntry(counter, d, System.currentTimeMillis());
    SETregister(se, d);
  }
  
  /*********************************
   ** SETregister(Document, long) **
   *********************************
   ** Method called by the SDS server or independent application to
   ** add a new document to the XSet search engine. Mainly a wrapper for
   ** generating the serviceID
   ** @param d The XML document to be added to the search engine
   ** @param counter The counter value assigned externally via a counter obj
   ** @exception XSetException for generalized errors
   **/
  public void SETregister(Document d, long counter) throws XSetException {
    //    System.err.println("Entering SETregister(Doc)");
    if (d == null) 
      throw new XSetException("Null document input to SETregister");
    ServiceEntry se = new ServiceEntry(counter, d, System.currentTimeMillis());
    SETregister(se, d);
  }

  /******************************************
   ** SETregister (ServiceEntry, Document) **
   ******************************************
   ** Method that takes in the ServiceEntry, and writes a stable version
   ** of it to disk first, before calling insertAndIndex to actually do
   ** the indexing.
   ** @param se The serviceentry object to be indexed
   ** @param d The DOM object corresponding to se.description
   ** @exception XSetException when file write fails
   **/
  void SETregister(ServiceEntry se, Document d) 
    throws XSetException {
    long documentNum = se.getNumID();

    //writing a file of the XML contents
    try {
      SETutils.writeServiceEntry(storeDir+File.separatorChar
				 +RecoveryMgr.dataDirName+File.separatorChar+
				 +documentNum+".xml", se);
    } catch (RecoveryException ioe) {
      throw new XSetException("XML file write failed: "+ioe.getMessage());
    }
    insertAndIndex(se, d, true);
  }

  /******************************************************
   ** insertAndIndex (ServiceEntry, Document, boolean) **
   ******************************************************
   ** The method which simply inserts or deletes the ServiceEntry into the
   ** indexing structure.
   ** @param se The serviceentry object to be indexed
   ** @param d The DOM object corresponding to se.description
   ** @param insdel Boolean value: true if insert, false if delete
   **/
  void insertAndIndex(ServiceEntry se, Document d, boolean insdel) {
    Treap rootTreap = null;
    Hashtable rootHash = null;
    Element root = d.getDocumentElement();
    String tagName = root.getTagName();
    int tType;
    String serviceID = se.getid();
    globalHashSID.put(serviceID, se);
    
    int i = SETutils.numChildren(root);
    // System.err.println("Register root:\n"+XMLHelper.SubtreeToString(root));
    if ((i > 1) || 
	(i == 1 && root.getFirstChild().getNodeType() == Node.ELEMENT_NODE)) {
      // HAS CHILDREN
      rootHash = (Hashtable) rootTagList.get(tagName+"&"+SETserver.HASHSTORE);
      //      System.err.println("Found root Hashtable");
      if (rootHash == null) {
	if (!insdel) {
	  //If trying to delete, missing rootTag, so return
	  return;
	}
	rootHash = new Hashtable();
	rootTagList.put(tagName+"&"+SETserver.HASHSTORE, rootHash);
      }
      addToHashSeq(rootHash, se, root, insdel);
    } else {
      // NO CHILDREN
      rootTreap = (Treap) rootTagList.get(tagName+"&"+SETserver.TREAPSTORE);
      //      System.err.println("Found root Treap");
      if (rootTreap == null) {
	if (!insdel) {
	  //If trying to delete, missing rootTag, so return
	  return;
	}
	rootTreap = new Treap(tagName);
	rootTagList.put(tagName+"&"+SETserver.TREAPSTORE, rootTreap);
      }
      addToTreap(rootTreap, se, root, insdel);
    }
    d = null;
  }
  
  /********************************************************
   ** addToTreap (Treap, ServiceEntry, Element, boolean) **
   ********************************************************
   ** Method called to take a serviceEntry, an element, and a treap,
   ** and insert/delete the element value into the treap, 
   ** registering the document
   **
   ** @param t The treap we're using to store the document hashtables under
   ** @param se The service entry that encapsulates document d
   ** @param elem The element in document d that we're filing under
   ** @param insdel Boolean: true if insert, false if delete
   **/
  private void addToTreap(Treap t, ServiceEntry se, 
			  Element elem, boolean insdel) {
    /* System.err.println("AddTreap: tag: "+ elem.getTagName()+" nodevalue: "
       +elem.getNodeValue()+
       " childnodes: "+elem.hasChildNodes()); */
    if (t == null || se == null || elem == null) {
      //If deleting, then return now.
      if (!insdel) return;
      System.err.println("Warning: null input values in addToTreap");
    }
    Hashtable hPtr;
    String tagValue;
    
    //    printTreap(t, "BEFORE");
    //    elem.normalize();
    Node child = null;
    if ((child = elem.getFirstChild()) == null 
	&& elem.getNodeType() == Node.ELEMENT_NODE) {
      tagValue = "";
    } else tagValue = child.getNodeValue();
    // look for the given value for this tag in the treap
    if ((hPtr = (Hashtable) t.lookup(tagValue)) == null) {
      // no entry found, make new hashtable and insert into the Treap
      // If deleting, then return now:
      if (!insdel) return;
      hPtr = new Hashtable();
      //System.err.println("adding new hashtable to treap for tag: "+tagValue);
      hPtr.put(se.getid(), new DocumentEntry(se, elem.getAttributes()));
      try {
	t.insert(tagValue, hPtr);
      } catch (DuplicateException de) {
	de.printStackTrace(System.err);
      }
      // add new hashtable to global list of hashtables (for cleaner)
      //	globalHashList.addElement(hPtr);
      // add a backpointer of the hashtable to the serviceEntry
      //      se.addHashtable(hPtr);
    } else {
      if (!insdel) hPtr.remove(se.getid());  //if delete, remove from hashtable
      else hPtr.put(se.getid(), new DocumentEntry(se, elem.getAttributes()));
      // add a backpointer of the hashtable to the serviceEntry
      //      se.addHashtable(hPtr);
      //      printTreap(t, "AFTER");
    }
  }
  
  /**************************************************************
   ** addToHashSeq (Hashtable, ServiceEntry, Element, boolean) **
   **************************************************************
   ** Method called to take a document, its root element, and a Hashtable,
   ** and traverse down the list of Hashtables, inserting/deleting the correct
   ** tags down the tree of Hashtables.
   ** @param h The current Hashtable we're looking at
   ** @param se The service entry that encapsulates document d
   ** @param root The tag in document d that we're filing under in t
   ** @param insdel True if inserting, false if deleting
   **/
  private void addToHashSeq(Hashtable currHash, ServiceEntry se, 
			    Element elem, boolean insdel) {
    //    System.err.println("Entering addToHashSeq(Hash,SE,Elem): "+
    //		       elem.getTagName().toString()+ " " + elem.getText());
    if (currHash == null || se == null || elem == null) {
      //If deleting, then just return
      if (!insdel) return;
      System.err.println("Warning: null inputs to addToHashSeq.");
    }
    Element ePtr = null;
    Vector ecPtr = null;
    String tagValue = null;
    
    Treap tPtr = null;
    Hashtable hPtr = null;
    
    ecPtr = SETutils.getChildren(elem);
    // take the list of children, and add them one by one
    int j = ecPtr.size();
    //    System.err.println("children #: "+j);
    //System.err.println("Add to hashseq:\n"+XMLHelper.SubtreeToString(elem));
    for(int i=0; i<j; i++) {
      //We are explicitly ignoring mixed content
      if (((Node)ecPtr.elementAt(i)).getNodeType() != Node.ELEMENT_NODE) {
	System.err.println("WARNING: mixed content ignored");
	continue;
      }
      ePtr = (Element) ecPtr.elementAt(i);
      //      System.err.println("child #: "+i+" "+ePtr.getText());
      //      System.err.println("  numElements: "+ePtr.numElements());
      
      //QUESTION: Should we do ePtr.normalize() here???
      //ANSWER: NO. Normalize only necessary after manual editing of text nodes
      int z = SETutils.numChildren(ePtr);
      if ((z > 1) || // child has children, => hashtable
	  (z==1 && ePtr.getFirstChild().getNodeType() == Node.ELEMENT_NODE)) { 
	hPtr = (Hashtable) currHash.get(ePtr.getTagName()+"&"+
					SETserver.HASHSTORE);
	if (hPtr == null) { // NOT FOUND
	  //If deleting, then skip to the next tag
	  if (!insdel) continue;
	  hPtr = new Hashtable();
	  currHash.put(ePtr.getTagName()+"&"+SETserver.HASHSTORE, hPtr);
	}
	addToHashSeq(hPtr, se, ePtr, insdel);
      } else {      // child has no children, => treap
	tPtr =(Treap)currHash.get(ePtr.getTagName()+"&"+SETserver.TREAPSTORE);
	if (tPtr == null) {
	  //If deleting, then skip to the next tag
	  if (!insdel) continue;
	  tPtr = new Treap(ePtr.getTagName().toString());
	  currHash.put(ePtr.getTagName()+"&"+SETserver.TREAPSTORE, tPtr);
	}
	addToTreap(tPtr, se, ePtr, insdel);
      }
    }
  }
  
  /*
    for each top level child tag C of "root" in the current query, 
    execute add Query C to each sub-treap that matches C.  if C is
    not in the tag treap, then create a new treap, add value of C to
    it, and add a reference to that XML doc to that node.  If C has
    no more children then the treap that C is associated with had better
    contain only values, NOT other treap pointers.
    
    As a result, each treap must have a few more things associated with
    it.  It needs to contain: 1. what each node's key's type is,
    whether it be tagName (or String), so that nodes point to other
    treaps, or else the keys are the actual values of the tag, and
    the nodes point to XML vectors. 2.  if the treaps are of the latter
    type, we must know what the types of the keys are, and there needs
    to be some comparison operator defined and associated with the
    type.  ie if type ==integer, then you're ok.  If type == String,
    we need the same associated string comparator operators.
  */
  
  /******************************
   ** SETderegister (Document) **
   ******************************
   ** Method called by the SDS server or independent application to
   ** remove a document from the SET search engine indices.
   **
   ** @param d The XML document to be removed from the search engine
   ** @exception XSetException for generalized errors
   **/
  public void SETderegister(Document d) throws XSetException {
    //    System.err.println("Entering SETdereg(Document d)");
    if (d==null) 
      throw new XSetException("Null document input to SETderegister.");
    String serviceID = null;
    try {
      serviceID = SETutils.hasher(d, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace(System.err);
      throw new XSetException("Missing SHA-1 hashing algorithm!");
    }
    ServiceEntry se = (ServiceEntry) globalHashSID.get(serviceID);
    delete(se);  
  }
  
  /**********************************
   ** SETderegister (ServiceEntry) **
   **********************************
   ** Method called by the XSet service to
   ** remove a document from the SET search engine indices.
   **
   ** @param se The XML ServiceEntry to be removed from the search engine
   ** @exception XSetException for generalized errors
   **/
  public void SETderegister(ServiceEntry se) throws XSetException {
    delete(se);  
  }
  
  /***************************
   ** delete (ServiceEntry) **
   ***************************
   ** Method called by SETderegister and the cleaner to remove
   ** a serviceEntry from the SET server.
   ** It looks up the hashtables pointing to it, and removes the
   ** references from them one by one.  It also removes the reference
   ** from the mter ServiceID to ServiceEntry hashtable.
   ** @param se The serviceEntry to be deleted.
   ** @exception XSetException for general runtime exceptions
   **/
  void delete(ServiceEntry se) throws XSetException {
    //System.err.println("Entering delete(ServiceEntry): "+se.toString());
    // get list of associated hashtables from the serviceEntry
    if (se == null) System.err.println("Null serviceEntry input to delete.");
    globalHashSID.remove(se.getid());

    Document d = null;
    try {
      ByteArrayInputStream bais = new ByteArrayInputStream(se.getXML().getBytes());
      parser.reset();
      parser.parse(new InputSource(bais));
      d = parser.getDocument();
    } catch (Exception e) {
      e.printStackTrace(System.err);
    }
    if (d == null) throw new XSetException("Error parsing XML to be deleted.");
    
    insertAndIndex(se, d, false);
    /**    Enumeration ePtr = se.getHashList().elements();
	   Hashtable hPtr = null;
	   while (ePtr.hasMoreElements()) {
	   // pull up the hashtable, and delete the se reference from it
	   hPtr = (Hashtable) ePtr.nextElement();
	   hPtr.remove(se.getid());
	   if (hPtr.isEmpty()) {
	   //NEED TO ADD CODE TO DELETE THE EMPTY HASHTABLE!!!!!!!!
	   }
	   }
    **/
  }

  /****************
   ** clearAll() **
   ****************
   ** Method that resets the SETserver objects and deletes all
   ** XML files in the data directory
   **/
  public void clearAll() {
    String fullpath = storeDir+File.separatorChar+
      RecoveryMgr.dataDirName;
    File f = new File(fullpath);
    if (!(f.exists() && f.isDirectory()))
      return;
    File temp;
    String [] files = f.list(new ExtensionFilter("xml"));
    if (files != null) {
      for (int i = 0; i < files.length; i++) {
	temp = new File(fullpath+File.separatorChar+files[i]);
	temp.delete();
      }
    }
    rootTagList = null;
    globalHashSID = null;
    parser = null;
    storeDir = null;
    if (runCleaner != null) {
      runCleaner.stop();
      runCleaner.destroy();
      runCleaner = null;
    }
  }

  /*************************************************************
               ALL DEBUGGING CODE FROM HERE ON
   *************************************************************/
  
  /***************************
   ** testCheckAttributes() **
   ***************************
   ** TESTER Method for checkAttributes
   **/
  public void testCheckAttributes() {
    Document d = null;
    try {
      parser.reset();
      parser.parse("e.xml");
      d = parser.getDocument();
    } catch (Exception e) { e.printStackTrace(System.err); return; }
    
    Element root = d.getDocumentElement();
    Vector ec = SETutils.getChildren(root);
    Element e = (Element) root.getFirstChild();
    
    // now need to create the DocumentEntry with attributes
    ServiceEntry se = null;
    try {
      se = new ServiceEntry(0,"", 0);
    } catch (XSetException xse) {
      xse.printStackTrace();
    }
    Element f = (Element) e.getNextSibling();
    DocumentEntry de = new DocumentEntry(se, f.getAttributes());
    Hashtable h = new Hashtable();
    h.put("a", de);
    Hashtable ans = checkAttributes(h, e);
    if (ans.isEmpty()) {
      System.out.println("No matches");
    } else {
      System.out.println("Matched!");
    }
  }
  
  /**********************
   ** testIntersect () **
   **********************
   ** TESTER Method for totalIntersect
   **/
  public void testIntersect () {
    HashVector hv = new HashVector();
    int i;
    Hashtable h = new Hashtable();
    for(i=1; i<10; i++) {
      h.put(String.valueOf(i), new Integer(i));
    }
    hv.addHashtable(h);
    h = new Hashtable();
    for (i=8; i<25; i++) {
      h.put(String.valueOf(i), new Integer(i));
    }
    hv.addHashtable(h);
    h = new Hashtable();
    for (i=3; i<15; i++) {
      h.put(String.valueOf(i), new Integer(i));
    }
    hv.addHashtable(h);
    Vector x = totalIntersect(hv);
    Enumeration e = x.elements();
    while (e.hasMoreElements()) {
      System.out.println(e.nextElement().toString());
    }
  }
  /**********************
   ** printSET(String) **
   **********************
   ** Debugging method used to print out the registration treap
   ** It uses the Treap.print()
   **/
  public void printSET(String roottagname) {
    //System.out.println("Entering printSET(Root Tag Name): "+roottagname);
    Hashtable roota = null;
    Treap rootb = null;
    Enumeration keys = rootTagList.keys();
    String k = null;
    while (keys.hasMoreElements()) {
      k = (String) keys.nextElement();
      System.out.println("Key in hashtable: "+k);
    }
    roota = (Hashtable) rootTagList.get(roottagname+"&"+SETserver.HASHSTORE);
    rootb = (Treap) rootTagList.get(roottagname+"&"+SETserver.TREAPSTORE);
    System.out.println("ENTERING PRINTER");
    if (roota != null) {
      printHash(roota, "");
    }
    if (rootb != null) {
      printTreap(rootb, "");
    }
  }
  
  /**********************************
   ** printHash(Hashtable, String) **
   **********************************
   ** Method used to recursively travel down the SET internal structure
   ** and print out information for debugging purposes
   ** @param h The hashtable to traverse/print
   ** @param offset The space offset to printout the tag with
   **/
  private void printHash(Hashtable h, String offset) {
    if (h == null) return;
    if (offset == null) offset = "";
    Enumeration keys = h.keys();
    String key;
    Object obj;
    while(keys.hasMoreElements()) {
      key = (String) keys.nextElement();
      System.out.println("Tag: "+offset+key);
      obj = h.get(key);
      if (obj instanceof Hashtable) {
	printHash((Hashtable) obj, offset+"   ");
      } else if (obj != null) {
	printTreap((Treap) obj, offset+"   ");
      }
    }
  }

  /********************************
   ** printTreap (Treap, String) **
   ********************************
   ** Method used to recursively travel down the SET internal structure
   ** and print out information for debugging purposes
   ** @param t The Treap to traverse/print
   **/
  private void printTreap(Treap t, String offset) {
    System.out.println("Printing Treap");
    Enumeration e = t.keys();
    String key;
    Hashtable h;
    while(e.hasMoreElements()) {
      key = (String) e.nextElement();
      h = (Hashtable) t.lookup(key);
      printH(h, offset, key);
    }
  }
  private void printH (Hashtable h, String offset, String key) {
    Enumeration hashe = h.keys();
    String hashkey;
    DocumentEntry de;
    while (hashe.hasMoreElements()) {
      hashkey = (String) hashe.nextElement();
      de = (DocumentEntry) h.get(hashkey);
      if (de == null || de.se == null) 
	System.out.println("EMPTY DOCUMENT_ENTRY");
      try {
	System.out.println(offset+"tag: "+key+" k: "+hashkey+ 
			   "\n"+offset+" id: "+de.se.getid()+"\n v: "+
			   de.se.getXML());
      } catch (SEntryFaultException sefe) {
	sefe.printStackTrace();
      }
    }
  }
}

