/**
 ** @author: Ben Y. Zhao (ravenben@cs.berkeley.edu)
 ** @version: CVS Version: $Revision: 1.7 $  Date: $Date: 2000/01/06 21:19:18 $
 ** (c) 2000 by Ben Yanbin Zhao and the Regents of the University of California
 ** <P>
 ** This is the wrapper class that surrounds the actual XML file.
 **/

package ninja.xset;
import org.w3c.dom.*;
import java.util.*;
import java.io.*;

public class ServiceEntry implements java.io.Serializable {
  private String id = null;
  private String description = null;
  private long lastHeartBeat = 0;
  private Vector hashtables = null;
  private long docnum;


  /**
   ** NOTE:  There could be problems in the hashing of identical XML
   ** but with different spacing into different service IDs
   ** Note that the hashing of the DOM object is actually hashing a
   ** canonical XML output that I generate.  Should fix this by
   ** adding a canonical XML transformer in SETutils.
   ** @param count The numerical id given to this XML doc by XSet
   ** @param d The DOM object that represents the XML string
   ** @param lastHeartBeat The last recorded heartbeat time for this doc
   ** @exception Thrown if the hasher algorithm does not exist
   **/
  public ServiceEntry(long count, Document d, long lastHeartBeat) 
    throws XSetException {
    this.docnum = count;
    this.description = XMLHelper.DocumentToString(d);
    try {
      this.id = SETutils.hasher(description, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace(System.err);
      throw new XSetException("Missing SHA-1 hash algorithm!");
    }
    this.lastHeartBeat = lastHeartBeat;
    this.hashtables = new Vector();
  }

  public ServiceEntry(long count, String xmltext, long lastHeartBeat) 
    throws XSetException {
    this.docnum = count;
    try {
      this.id = SETutils.hasher(xmltext, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace(System.err);
      throw new XSetException("Missing SHA-1 hash algorithm!");
    }
    this.description = xmltext;
    this.lastHeartBeat = lastHeartBeat;
    this.hashtables = new Vector();
  }

  public ServiceEntry(String stringid, long count, String xmltext, 
		      long lastHeartBeat) throws XSetException {
    this.id = stringid;
    this.docnum = count;
    try {
      this.id = SETutils.hasher(xmltext, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace(System.err);
      throw new XSetException("Missing SHA-1 hash algorithm!");
    }
    this.description = xmltext;
    this.lastHeartBeat = lastHeartBeat;
    this.hashtables = new Vector();
  }
  
  public void updateTime(long newHeartBeat) {
    lastHeartBeat = newHeartBeat;
  }
  public long getHeartBeat() {
    return lastHeartBeat;
  }
  public String getid() {
    return id;
  }

  /*********************
   ** long getNumID() **
   *********************
   ** Accessor method for the numerical sequence ID
   ** @return The sequencer ID as a long
   **/
  public long getNumID() {
    return docnum;
  }

  /*****************************
   ** addHashtable(Hashtable) **
   *****************************
   ** Method to add hashtable to the list of backpointers
   ** @param h the hashtable to be added
   **/
  //hashtable backpointers eliminated for memory efficiency
  /*  public void addHashtable(Hashtable h) {
    hashtables.addElement(h);
    }*/

  /**************************
   ** Vector getHashList() **
   **************************
   ** method to return the list of backpointer hashtables
   ** @return Vector of hashtable backpointers
   **/
  //hashtable backpointers eliminated for memory efficiency
  /*  public Vector getHashList() {
    return hashtables;
    }*/

  public String toString() {
    String out = "";
    out = out+"XML:\n"+description+"\nLast Beat:\n"+lastHeartBeat+"\n";
    return out;
  }

  /*********************
   ** String getXML() **
   *********************
   ** Accessor method to the XML description
   ** @return the XML description as a String
   ** @exception SEntryFaultException indicates XML has been paged out
   **/
  public String getXML() throws SEntryFaultException {
    if (description == null)
      throw new SEntryFaultException("ServiceEntry XML string paged out or not set!");
    return description;
  }

  /***************************
   ** String getXML(String) **
   ***************************
   ** Accessor method that automatically paged in documents
   ** @return the XML description as a String
   **/
  public String getXML(String datafilePath) {
    if (description == null) 
      pageinXML(datafilePath);
    return description;
  }

  /*********************
   ** void evictXML() **
   *********************
   ** method to evict the XML document from memory for memory savings
   ** This is done by setting the description variable to null
   ** The document can be brought back on a read fault using pageinXML
   **/
  public void evictXML() {
    description = null;
  }

  /*************************************
   ** void pageinXML(String storeDir) **
   *************************************
   ** Method to bring back the XML document text value into memory.
   ** Looks for a .xml file with the docnum as the name
   **/
  public void pageinXML(String storeDir) {
    File xmlfile = null;
    xmlfile = new File(storeDir+"/"+docnum+".xml");
    if (xmlfile == null) return;

    StringBuffer sb = null;
    BufferedReader br = null;
    String temp;
    try {
      sb = new StringBuffer();
      br = new BufferedReader(new FileReader(xmlfile));
      while ((temp = br.readLine()) == null) {
	sb.append(temp);
      }
      description = sb.toString();
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }
}
