/**
 ** @author: Ben Y. Zhao (ravenben@cs.berkeley.edu)
 ** @version: CVS Version: $Revision: 1.15 $  Date: $Date: 2000/01/06 21:19:18 $
 ** (c) 2000 by Ben Yanbin Zhao and the Regents of the University of California
 **/

package ninja.xset;

import java.io.*;
import java.util.*;
import java.rmi.RemoteException;
import org.w3c.dom.*;
import org.xml.sax.InputSource;
import java.rmi.*;
import java.rmi.server.*;

/**
 ** Class that makes the SETserver an ISpace service
 ** It offers methods to addfiles to the XSet search engine, clear the XSet 
 ** contents, and to query XSet
 **
 ** As an ISpace service, the correct entry in ispace.cfg is
 **    <tt>ninja.xset.XSetService</tt>
 ** The service is originally empty, and files can be added directory
 ** by directory by calling XSetAddFiles(path)
 **/
public class XSetService extends UnicastRemoteObject implements XSetServiceIF {

  //CLASS fields
  private SETserver server;
  private String storeDir = null;
  private int initBufferLogSize = 20;
  
  DOMParser parser;
  RecoveryMgr recovMgr;

  /**
   ** Constructor, just creates an SETserver object
   **/
  public XSetService() throws RemoteException {
    parser = new DOMParser();
  }
  
  /*******************************
   ** init(iSpaceServiceConfig) **
   *******************************
   ** Initialization: checks the arguments and prints out 
   ** informative message. <BR>
   ** Now needs to check local filesystem for backup, and restore previous 
   ** database contents from checkpoint.
   **/
   protected void init(String args[]) {
    boolean cleanBoot = false;

    if (args == null || args.length == 0) {
      System.err.println("Usage: ninja.xset.XSetService [-b] storeDir");
      System.exit(-1);
    }
    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-b")) cleanBoot = true;
      else storeDir = args[i];
    }
    if (storeDir == null) {
      System.err.println("Missing argument: initialStoragePath");
      System.exit(-1);
    }
    try {
      recovMgr = new RecoveryMgr(storeDir, initBufferLogSize);
    } catch (RecoveryException re) {
      System.out.println("Error: RecoveryMgr buffersize out of range.");
      return;
    }

    if (!cleanBoot) {
      try {
	server = recovMgr.restore();
      } catch (RecoveryException re) {
	System.out.println("Error: checkpoint recovery failed.\n"+
			   "Reinitializing SETserver.");
	server = new SETserver(storeDir);
      }
    } else {
      server = new SETserver(storeDir);
    }
    System.out.println("XSetService: started.");
  }

  protected void destroy() {
    System.out.println("XSetService: stopped.");
    server = null;
  }

  /******************************
   ** String getDatafilePath() **
   ******************************
   ** Method to grab the full path of the datafiles, which is
   ** the same as the storeDir + the data directory
   ** @return The full path to the XML data files
   **/
  public String getStoreDir() {
    return (storeDir +File.separatorChar+recovMgr.dataDirName);
  }

  /**************************
   ** AddDocument (String) **
   **************************
   ** Method used to add a single document to the XSet search engine index
   ** @param xmlstring the document to be added
   ** @exception RemoteException as a remote service should
   **/
  public void AddDocument(String xmlstring) 
    throws java.rmi.RemoteException {
    long counter = server.count.getNextCount();
    try {
      recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPEINSERT,
						 counter, LogEntry.LOGBEGIN));
      server.SETregister(xmlstring, counter);
      recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPEINSERT,
						 counter, LogEntry.LOGEND));
    } catch (XSetException xse) {
      throw new java.rmi.RemoteException(xse.getMessage());
    } catch (CheckpointException cpe) {
      throw new java.rmi.RemoteException("Checkpoint Exception"+cpe.getMessage());
    }
  }
 
  /***********************
   ** AddFiles (String) **
   ***********************
   ** Method used to add files to the XSet search engine index
   ** @param path String that encodes the full path of the directory
   ** that contains the files to be indexed
   ** @exception RemoteException as demanded by RMI
   **/
  public void AddFiles(String path) throws java.rmi.RemoteException {
    System.err.println("AddFiles request received!");
    File dir = new File(path);
    String [] xmlFiles = null;
    Document d = null;
    
    long counter;
    if ((xmlFiles = dir.list(new ExtensionFilter("xml"))) == null) {
      System.err.println("Directory not found or inaccessible!");
    }
    int i = 0;
    try {
      parser.reset();
      for (i = 0; i < xmlFiles.length; i++) {
	System.err.println("File: "+xmlFiles[i]+ " loaded.");
	parser.parse(path+File.separatorChar+xmlFiles[i]);
	d = parser.getDocument();
	if (d != null) {
	  counter = server.count.getNextCount();
	  recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPEINSERT,
						     counter, LogEntry.LOGBEGIN));
	  server.SETregister(d, counter);
	  recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPEINSERT,
						     counter, LogEntry.LOGEND));
	} else System.err.println("File failed to parse: "+
				  xmlFiles[i] +", ignored.");
      }
    } catch (FileNotFoundException fnfe) {
      fnfe.printStackTrace(System.err);
      throw new RemoteException("Query file not found!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      throw new RemoteException("Parsing ERROR on: "+xmlFiles[i]);
    }
  }

  /*********************
   ** AddFile(String) **
   *********************
   ** Method to add a single file to the XSet dataset
   ** @param filename String encoding of the filename including full path
   ** @exception RemoteException if file not found or error in IO.
   **/
  public void AddFile(String filename) throws java.rmi.RemoteException {
    //    System.err.println("AddFile request received!");
    try {
      parser.reset();
      parser.parse(filename);
      Document d = parser.getDocument();
      long sequenceID = server.count.getNextCount();
      if (d != null) {
	recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPEINSERT,
						   sequenceID, LogEntry.LOGBEGIN));
	server.SETregister(d, sequenceID);
	recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPEINSERT,
						   sequenceID, LogEntry.LOGEND));
	//NOTE: IF SETregister fails, then logEntry made pointing to no FILE
      } else 
	throw new java.rmi.RemoteException("Parsing ERROR on: "+filename);
    } catch (FileNotFoundException fnfe) {
      fnfe.printStackTrace(System.err);
      throw new RemoteException("Query file not found!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      throw new RemoteException("Parsing ERROR on: "+filename);
    }
  }

  /****************
   ** ClearAll() **
   ****************
   ** Method to clear the XSet index, in reality it just creates
   ** a new instance of the SETserver object
   **/
  public void ClearAll() throws java.rmi.RemoteException {
    System.err.println("ClearAll request received!");
    //RESET recoveryManager
    try {
      recovMgr = new RecoveryMgr(storeDir, initBufferLogSize);
    } catch (RecoveryException re) {
      re.printStackTrace(System.err);
    }
    //RESET SETserver object
    if (server != null) server.clearAll();
    server = new SETserver(storeDir);
    System.err.println("XSet cleared.");
  }

  /*****************************
   ** Vector QuerySE (String) **
   *****************************
   ** Method to query the XSet server
   ** @param xmlQuery is the query in the form of a text string encoded xml
   ** document.  The query must be well-formed, but not contain a reference
   ** to any DTDs
   ** @return Vector of ServiceEntry objects, each wrapped around XML Document
   **/
  public Vector QuerySE(String xmlQuery) throws java.rmi.RemoteException {
    System.err.println("Query request received!");
    parser.reset();
    Document d;
    Vector results = null;
    try {
      ByteArrayInputStream bais =new ByteArrayInputStream(xmlQuery.getBytes());
      parser.parse(new InputSource(bais));
      d = parser.getDocument();
      if (d != null) results = server.querySET(d);
      else throw new java.rmi.RemoteException("Query XML parse error!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      System.err.println("Error parsing query string: \n"+xmlQuery);
      results = new Vector();
    }
    return results;
  }

  /**************************
   ** Vector Query(String) **
   **************************
   ** Method that queries the XSet server and returns vector of XML documents
   ** @param xmlQuery is the query in string encoded xml format, which must be well-formed, and should not contain a reference to any DTDs
   ** @return Vector of XML documents in string form
   **/
  public Vector Query(String xmlQuery) throws java.rmi.RemoteException {
    Vector results = QuerySE(xmlQuery);
    //Results will not be null. If no match, then empty vector will be returned
    if (results == null) {
      System.err.println("ERROR: null results returned!!");
      return new Vector();
    }
    if (results.size() == 0) return results;
    Vector xml = new Vector();
    Enumeration e = results.elements();
    ServiceEntry se;
    while (e.hasMoreElements()) {
      se = (ServiceEntry) e.nextElement();
      xml.addElement(se.getXML(storeDir+File.separatorChar+
			       RecoveryMgr.dataDirName));
    }
    System.out.println(xml.size());
    //    System.out.println(xml);
    return xml;
  }
  
  /***********************************
   ** Vector Query (String, Vector) **
   ***********************************
   ** Method to query the XSet server, and return only those fields requested
   ** <P>
   ** This returned data structure is rather complicated.  It's a vector of
   ** hashtables, one hashtable for each matching XML document found. For
   ** each of these matches, the hashtable hashes each of the requested
   ** tag names to a vector.  The vector contains all the string values
   ** which occur in that tagname in the currently matched XML document.
   ** So it's a Vector of Hashtables of Vectors.
   ** <P>
   ** @param xmlQuery, the xml query in the form of a string encoded XML
   ** @param tagVector, a vector containing String names of requested tags
   ** @return Vector of hashtables that map each requested tag name to its value
   **/
  public Vector Query(String xmlQuery, Vector tagVector) 
    throws java.rmi.RemoteException {
    
    System.err.println("Query request received!\n"+xmlQuery);
    parser.reset();
    Document d;
    Vector results = null;
    try {
      ByteArrayInputStream bais =new ByteArrayInputStream(xmlQuery.getBytes());
      parser.parse(new InputSource(bais));
      d = parser.getDocument();
      if (d != null) results = server.querySET(d, tagVector);
      else throw new java.rmi.RemoteException("Query XML parse error!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      System.err.println("Error processing query: \n"+xmlQuery);
      results = new Vector();
    }
    return results;
  }

  /*********************************
   ** EnumerateTag (String []) ***
   **********************************
   ** Method that takes in a path to a subpath down to a tag in the form of an
   ** array, and returns an enumeration of the existing values in the treap for
   ** that tag.  Just a simple wrapper around SETserver.enumerateTag()
   ** @param pathArray - the string array of tag names in the path to the tag
   ** @return Vector of unique existing values that exist in the dataset
   **/
  public Vector EnumerateTag(String [] pathArray) 
    throws java.rmi.RemoteException {
    Enumeration e;
    try {
      e = server.enumerateTag(pathArray);
    } catch (XSetException xse) {
      throw new RemoteException(xse.getMessage());
    }
    Vector newV = new Vector();
    while (e.hasMoreElements())
      newV.addElement(e.nextElement());
    return newV;
  }
  
  /********************************
   ** RemoveDocument(Document) **
   *********************************
   ** The method that removes a document from the index structure
   ** @param d The org.w3c.dom document that is to be removed
   **/
  public void RemoveDocument(Document d) throws java.rmi.RemoteException {
    System.err.println("RemoveDocument (Document) request received!");
    if (d==null) 
      throw new java.rmi.RemoteException("Null document input to removeDocument.");
    String serviceID = null;
    try {
      serviceID = SETutils.hasher(d, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace(System.err);
      throw new java.rmi.RemoteException("Missing SHA-1 hashing algorithm!");
    }
    ServiceEntry se = (ServiceEntry) server.globalHashSID.get(serviceID);
    long sequenceID = se.getNumID();
    try {
      recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPEDELETE,
						 sequenceID, LogEntry.LOGBEGIN));
      server.SETderegister(se);
      recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPEDELETE,
						 sequenceID, LogEntry.LOGEND));
    } catch (XSetException xse) {
      throw new java.rmi.RemoteException(xse.getMessage());
    } catch (CheckpointException ce) {
      throw new java.rmi.RemoteException(ce.getMessage());
    }
  }

  /*******************************
   *** RemoveDocument(String) **
   ********************************
   ** The method that takes in a XML string, and tries to remove it from the
   ** index if it exists
   ** @param xml The XML string representing the XML document to be removed
   **/
  public void RemoveDocument(String xml) throws java.rmi.RemoteException {
    System.err.println("RemoveDocument (String) request received!");
    parser.reset();
    Document d;
    try {
      ByteArrayInputStream bais =new ByteArrayInputStream(xml.getBytes());
      parser.parse(new InputSource(bais));
      d = parser.getDocument();
      if (d != null) {
	RemoveDocument(d);
      } else 
	throw new java.rmi.RemoteException("Parsing error on XML String!");
    } catch (Exception e) {
      e.printStackTrace(System.err);
      throw new java.rmi.RemoteException(e.getMessage());
    }
  }

  /***************************
   ** boolean Checkpoint () **
   ***************************
   ** The method to do synchronous checkpointing.<BR>
   ** It generates a new sequenceID for the checkpoint log entry
   ** @return whether the checkpointing failed (-1) or succeeded (0)
   ** @exception RemoteException which encapsulates any server side exceptions
   **/
  public boolean Checkpoint() throws java.rmi.RemoteException {
    boolean b;

    long sequenceID = server.count.getNextCount();
    try {
      recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPECHECKPOINT,
						 sequenceID, LogEntry.LOGBEGIN));
      b = recovMgr.checkpoint(server);
      recovMgr.writeLog(RecoveryMgr.makeLogEntry(LogEntry.LOGTYPECHECKPOINT,
						 sequenceID, LogEntry.LOGEND));
    } catch (CheckpointException re) {
      throw new java.rmi.RemoteException(re.getMessage());
    }
    return b;
  }

  /**************************
   ** boolean ForceSync () **
   **************************
   ** The method to force log buffers to be written out to disk.<BR>
   ** @exception RemoteException which encapsulates any server side exceptions
   **/
  public void ForceSync() throws java.rmi.RemoteException {
    try {
      recovMgr.forceSync();
    } catch (CheckpointException ce) {
      throw new java.rmi.RemoteException(ce.getMessage());
    }
  }

  /**********************************
   ** int Shutdown(boolean backup) **
   **********************************
   ** The method for shutting down a XSetService
   ** @param backup - whether to checkpoint before shutdown
   ** @return success or failure of the checkpointing before shutdown
   **/
  public int Shutdown(boolean backup) throws java.rmi.RemoteException {
    int b = 0;
    if (backup) {
      try {
	if (!recovMgr.checkpoint(server)) b = -1;
      } catch (CheckpointException re) {
	throw new java.rmi.RemoteException(re.getMessage());
      }
    }
    //ADD CODE FOR SHUTTING DOWN
    return b;
  }

    public static void main (String args[]) {
	try {
	    XSetService srv = new XSetService ();
	    srv.init(args);
	    Naming.rebind("//localhost/XSet", srv);
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }
}
